/**
  *  \file planetscentral/console/context.cc
  *  \brief Console: Context Handling
  */

#include <stdexcept>
#include <iostream>
#include <sys/time.h>
#include <time.h>
#include "planetscentral/console/context.h"
#include "cpluslib/format.h"
#include "util/misc.h"

/** Constructor.
    \param name User-visible name. */
ConnectionContext::ConnectionContext(string_t name)
    : name(name),
      host("127.0.0.1"),
      port("53535")
{ }

void
ConnectionContext::enter()
{
    if (!client.isConnected()) {
        std::cout << "(connecting to " << host << ":" << port << "...)\n";
        client.connect(host, port);
    }
}

void
ConnectionContext::leave()
{ }

IntValue*
ConnectionContext::processCommand(StringList_t& cmd)
{
    if (cmd[0] == "repeat") {
        // Process command repeatedly, for benchmarking
        if (cmd.size() < 3) {
            throw std::runtime_error("Too few arguments");
        }
        int32_t n;
        if (!stringValue(cmd[1], n) || n <= 0) {
            throw std::runtime_error("Expecting number");
        }
        if (client.maybeReconnect(host, port)) {
            std::cout << "(auto-reconnecting to " << host << ":" << port << "...)\n";
        }
        struct timeval beg, end;
        gettimeofday(&beg, 0);
        for (int32_t i = 0; i < n; ++i) {
            IntDataSegment dseg;
            for (uint32_t x = 2; x < cmd.size(); ++x) {
                dseg.pushNew(makeStringValue(cmd[x]));
            }
            client.callVoid(dseg);
        }
        gettimeofday(&end, 0);

        int32_t seconds = end.tv_sec - beg.tv_sec;
        int32_t usecs   = end.tv_usec - beg.tv_usec;
        if (usecs < 0) {
            seconds--;
            usecs += 1000000;
        }
        int32_t msecs = usecs / 1000;
        
        return makeStringValue(format("%d.%03d seconds (=%d ms per iteration)",
                                      seconds, msecs, (1000*seconds+msecs)/n));
    } else if (cmd[0] == "reconnect" || cmd[0] == "reset") {
        // Reconnect
        std::cout << "(reconnecting to " << host << ":" << port << "...)\n";
        client.connect(host, port);
        return makeStringValue("OK");
    } else {
        // Process command directly
        uint32_t first = (cmd[0] == "exec" ? 1 : 0);
        IntDataSegment dseg;
        for (uint32_t x = first; x < cmd.size(); ++x) {
            dseg.pushNew(makeStringValue(cmd[x]));
        }
        if (client.maybeReconnect(host, port)) {
            std::cout << "(auto-reconnecting to " << host << ":" << port << "...)\n";
        }
        return client.call(dseg);
    }
}

string_t
ConnectionContext::getPrefix()
{
    return name;
}

/** Check configuration.
    \param prefix This context's configuration prefix
    \param key Received key
    \param value Received value
    \retval true Key was recognized
    \retval false Key was not recognized */
bool
ConnectionContext::checkConfig(string_t prefix, string_t key, string_t value)
{
    if (key == prefix + ".HOST") {
        this->host = value;
        return true;
    }
    if (key == prefix + ".PORT") {
        this->port = value;
        return true;
    }
    return false;
}
