/**
  *  \file planetscentral/monitor.cc
  *  \brief Service Monitor
  *
  *  This is a standalone service monitor, reporting status per HTTP. 
  *  It can be run under c2logger, and takes configuration via
  *  C2CONFIG.
  */

#include <sys/socket.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/select.h>
#include <sys/time.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <ctype.h>
#include <errno.h>
#include "cpluslib/string.h"
#include "cpluslib/types.h"
#include "server/request.h"
#include "util/misc.h"
#include "server/config.h"
#include "server/net.h"

enum { TIME = 10 };

static string_t arg_listen_host = "0.0.0.0";
static string_t arg_listen_port = "8080";
static string_t arg_template_file = "planetscentral/monitor.html";
static int arg_interval = 10;

static string_t arg_router_host = "127.0.0.1";
static string_t arg_router_port = "9999";

static string_t arg_www_host = "127.0.0.1";
static string_t arg_www_port = "80";

static string_t arg_redis_host = "127.0.0.1";
static string_t arg_redis_port = "6379";

static string_t arg_file_host = "127.0.0.1";
static string_t arg_file_port = "9998";

static string_t arg_hostfile_host = "127.0.0.1";
static string_t arg_hostfile_port = "7776";

static string_t arg_host_host = "127.0.0.1";
static string_t arg_host_port = "7777";

static string_t arg_mailout_host = "127.0.0.1";
static string_t arg_mailout_port = "21212";

static string_t arg_talk_host = "127.0.0.1";
static string_t arg_talk_port = "5555";

static string_t arg_pop3_error;

static string_t html_template;
static string_t html_status;
static time_t html_template_time;
static time_t html_status_time;
static time_t now;

enum State {
    isOK,
    isBroken,
    isDown
};

static void
updateTemplate()
{
    /* Do we have to reload the template? */
    struct stat st;
    if (stat(arg_template_file.c_str(), &st) < 0) {
        html_template = strerror(errno);
        return;
    }
    if (st.st_mtime == html_template_time) {
        return;
    }
    html_template_time = st.st_mtime;

    /* Yes, we must update */
    int fd = open(arg_template_file.c_str(), O_RDONLY);
    if (fd < 0) {
        html_template = strerror(errno);
        return;
    }

    /* Read it */
    Request req;
    char* line;
    html_template.clear();
    while (1) {
        if ((line = req.readLine()) != 0) {
            html_template += line;
            html_template += "\r\n";
        } else if (req.readMore(fd) != Request::sSuccess) {
            break;
        }
    }

    /* Done. */
    close(fd);
    printf("c2monitor: loaded template\n");
}

static void
addStatus(const char* service, State st)
{
    html_status += "<tr><td class=\"service\">";
    html_status += service;
    html_status += "</td>";
    if (st == isOK) {
        html_status += "<td class=\"running\">running</td></tr>\n";
        printf("    %s is running\n", service);
    } else if (st == isBroken) {
        html_status += "<td class=\"suspended\" title=\"service has failed\">broken</td></tr>\n";
        printf("    %s is broken\n", service);
    } else {
        html_status += "<td class=\"down\" title=\"service is unexpectedly down\">down</td></tr>\n";
        printf("    %s is down\n", service);
    }
}

static bool
callServer(string_t host, string_t port, string_t req, string_t& reply, bool useShutdown)
{
    /* Clear reply, just in case */
    reply.clear();

    /* Special case: if host is 0.0.0.0, connect to localhost */
    if (host.find_first_not_of("0.") == host.npos) {
        host = "127.0.0.1";
    }

    /* Avoid being stupid */
    if (port == arg_listen_port) {
        return false;
    }

    /* Connect */
    int sock = TimedConnect(10).withAddress(host, port);
    if (sock < 0) {
        return false;
    }

    /* For the time being, make it blocking again */
    fcntl(sock, F_SETFL, 0);

    /* Send some data */
    write(sock, req.data(), req.size());
    if (useShutdown)
        shutdown(sock, SHUT_WR);

    /* Read reply */
    Request reqb;
    while (1) {
        const char* line;
        if ((line = reqb.readLine()) != 0) {
            if (reply.size() < 4096) {
                reply += line;
                reply += "\r\n";
            }
        } else if (reqb.readMore(sock) != Request::sSuccess) {
            break;
        }
    }
    close(sock);
    return true;
}

static State
checkStatusFile(const string_t& name)
{
    int fd = open(name.c_str(), O_RDONLY);
    if (fd < 0) {
        /* file does not exist: service must be down */
        return isDown;
    }

    /* Read content */
    char buffer[20];
    struct stat st;
    int readresult = read(fd, buffer, sizeof(buffer)-1);
    int statresult = fstat(fd, &st);
    close(fd);

    if (readresult < 0 || readresult >= int(sizeof(buffer)) || statresult < 0) {
        /* I/O error: system is broken somehow */
        return isDown;
    }

    if (time(0) - st.st_mtime > 3600) {
        /* Status file has not been changed in an hour: service must be down */
        return isDown;
    }

    if (readresult == 0) {
        /* File is empty. This can happen during the small instant of time the
           service updates the file. So assume for the benefit of the doubt that
           the system is OK. */
        return isOK;
    }

    buffer[readresult] = 0;
    char* endptr;
    long badness = strtol(buffer, &endptr, 10);
    if (*endptr != 0 || badness < 0) {
        /* Syntax error */
        return isBroken;
    }

    if (badness > 1) {
        /* More than one bad cycle: broken */
        return isBroken;
    } else {
        /* Everything OK */
        return isOK;
    }
}

static void
updateStatus()
{
    /* Don't recheck if it is recent */
    if (html_status.size() && difftime(now, html_status_time) < arg_interval)
        return;

    /* Check */
    html_status_time = now;
    html_status.clear();
    printf("c2monitor: checking status:\n");

    /* WWW */
    string_t result;
    if (callServer(arg_www_host, arg_www_port, "GET / HTTP/1.0\r\nHost: 127.0.0.1\r\n\r\n", result, true)) {
        if (result.size() > 12 && result.compare(0, 6, "HTTP/1", 6) == 0 && result.compare(9, 3, "200", 3) == 0) {
            addStatus("Web Server", isOK);
        } else {
            addStatus("Web Server", isBroken);
        }
    } else {
        addStatus("Web Server", isDown);
    }

    /* c2router */
    if (callServer(arg_router_host, arg_router_port, "LIST", result, true)) {
        addStatus("PCC2 Web", isOK);
    } else {
        addStatus("PCC2 Web", isDown);
    }

    /* redis */
    if (callServer(arg_redis_host, arg_redis_port, "PING\r\nQUIT\r\n", result, false)) {
        if (result.size() >= 5 && result.compare(0, 5, "+PONG", 5) == 0) {
            addStatus("Database", isOK);
        } else {
            addStatus("Database", isBroken);
        }
    } else {
        addStatus("Database", isDown);
    }

    /* c2file */
    if (callServer(arg_file_host, arg_file_port, "PING\n", result, true)) {
        if (result.size() > 5 && result.find("PONG") != result.npos) {
            addStatus("User File Server", isOK);
        } else {
            addStatus("User File Server", isBroken);
        }
    } else {
        addStatus("User File Server", isDown);
    }

    /* c2file */
    if (callServer(arg_hostfile_host, arg_hostfile_port, "PING\n", result, true)) {
        if (result.size() > 5 && result.find("PONG") != result.npos) {
            addStatus("Host File Server", isOK);
        } else {
            addStatus("Host File Server", isBroken);
        }
    } else {
        addStatus("Host File Server", isDown);
    }

    /* c2host */
    if (callServer(arg_host_host, arg_host_port, "PING\n", result, true)) {
        if (result.size() > 5 && result.find("PONG") != result.npos) {
            addStatus("Host Manager", isOK);
        } else {
            addStatus("Host Manager", isBroken);
        }
    } else {
        addStatus("Host Manager", isDown);
    }

    /* c2mailout */
    if (callServer(arg_mailout_host, arg_mailout_port, "PING\n", result, true)) {
        if (result.size() > 5 && result.find("PONG") != result.npos) {
            addStatus("Mail Manager", isOK);
        } else {
            addStatus("Mail Manager", isBroken);
        }
    } else {
        addStatus("Mail Manager", isDown);
    }

    /* c2talk */
    if (callServer(arg_talk_host, arg_talk_port, "PING\n", result, true)) {
        if (result.size() > 5 && result.find("PONG") != result.npos) {
            addStatus("Forum", isOK);
        } else {
            addStatus("Forum", isBroken);
        }
    } else {
        addStatus("Forum", isDown);
    }

    /* pop3 */
    if (!arg_pop3_error.empty()) {
        addStatus("Mail Fetch", checkStatusFile(arg_pop3_error));
    }
}

static void
writeTemplate(int fd, const char* tpl)
{
    char buffer[128];
    while (const char* p = strstr(tpl, "$(")) {
        if (p != tpl)
            timedFullWrite(fd, tpl, p-tpl, TIME);
        if (strncmp(p, "$(STATUS)", 9) == 0) {
            timedFullWrite(fd, html_status.data(), html_status.size(), TIME);
            tpl = p+9;
        } else if (strncmp(p, "$(CHECKTIME)", 12) == 0) {
            strftime(buffer, sizeof(buffer), "%d/%b/%Y %H:%M:%S GMT", gmtime(&html_status_time));
            timedWriteStr(fd, buffer, TIME);
            tpl = p+12;
        } else if (strncmp(p, "$(SERVERTIME)", 13) == 0) {
            strftime(buffer, sizeof(buffer), "%d/%b/%Y %H:%M:%S GMT", gmtime(&now));
            timedWriteStr(fd, buffer, TIME);
            tpl = p+13;
        } else if (strncmp(p, "$(CHECKTIMENUM)", 15) == 0) {
            sprintf(buffer, "%lld", (long long) html_status_time);
            timedWriteStr(fd, buffer, TIME);
            tpl = p+15;
        } else if (strncmp(p, "$(SERVERTIMENUM)", 16) == 0) {
            sprintf(buffer, "%lld", (long long) now);
            timedWriteStr(fd, buffer, TIME);
            tpl = p+16;
        } else {
            timedWriteStr(fd, "FIXME: template error", TIME);
            tpl = p+1;
        }
    }
    timedWriteStr(fd, tpl, TIME);
}

static void
handleRequest(int fd)
{
    Request req;

    /* Read command */
    char* command;
    while ((command = req.readLine()) == 0) {
        if (req.readMore(fd) != Request::sSuccess) {
            writeStr(fd, "400 Bad request\n");
            return;
        }
    }

    /* Find command verb */
    string_t verb;
    while (*command != 0 && *command != ' ' && *command != '\t') {
        verb += toupper(uint8_t(*command++));
    }

    /* Read remainder of request */
    while (1) {
        char* line;
        while ((line = req.readLine()) == 0) {
            if (req.readMore(fd) != Request::sSuccess) {
                writeStr(fd, "400 Bad request\n");
                return;
            }
        }
        if (*line == 0) {
            break;
        }
    }

    /* Check command verb */
    if (verb == "GET") {
        updateTemplate();
        updateStatus();
        writeStr(fd,
                 "HTTP/1.0 200 OK\r\n"
                 "Connection: close\r\n"
                 "Content-Type: text/html\r\n"
                 "Pragma: no-cache\r\n"
                 "Cache-Control: no-cache\r\n"
                 "\r\n");
        writeTemplate(fd, html_template.c_str());
    } else {
        timedWriteStr(fd, "400 Bad request\n", TIME);
    }
}

static void
processConfig(string_t key, string_t value)
{
    if (key == "MONITOR.HOST") {
        arg_listen_host = value;
    } else if (key == "MONITOR.PORT") {
        arg_listen_port = value;
    } else if (key == "MONITOR.TEMPLATE") {
        arg_template_file = value;
    } else if (key == "MONITOR.INTERVAL") {
        stringValue(value, arg_interval);
    } else if (key == "ROUTER.HOST") {
        arg_router_host = value;
    } else if (key == "ROUTER.PORT") {
        arg_router_port = value;
    } else if (key == "WWW.HOST") {
        arg_www_host = value;
    } else if (key == "WWW.PORT") {
        arg_www_port = value;
    } else if (key == "REDIS.HOST") {
        arg_redis_host = value;
    } else if (key == "REDIS.PORT") {
        arg_redis_port = value;
    } else if (key == "FILE.HOST") {
        arg_file_host = value;
    } else if (key == "FILE.PORT") {
        arg_file_port = value;
    } else if (key == "HOSTFILE.HOST") {
        arg_hostfile_host = value;
    } else if (key == "HOSTFILE.PORT") {
        arg_hostfile_port = value;
    } else if (key == "HOST.HOST") {
        arg_host_host = value;
    } else if (key == "HOST.PORT") {
        arg_host_port = value;
    } else if (key == "MAILOUT.HOST") {
        arg_mailout_host = value;
    } else if (key == "MAILOUT.PORT") {
        arg_mailout_port = value;
    } else if (key == "TALK.HOST") {
        arg_talk_host = value;
    } else if (key == "TALK.PORT") {
        arg_talk_port = value;
    } else if (key == "POP3.ERROR") {
        arg_pop3_error = value;
    }
}

int main()
{
    /* Unbuffer stdout */
    setvbuf(stdout, 0, _IOLBF, 0);

    /* Configuration */
    loadConfiguration(processConfig);

    /* Listen */
    int sock = Bind().withAddress(arg_listen_host, arg_listen_port);
    if (sock < 0) {
        fprintf(stderr, "unable to bind socket\n");
        return 1;
    }

    /* OK, main loop */
    printf("c2monitor: listening...\n");
    signal(SIGPIPE, SIG_IGN);
    fcntl(sock, F_SETFD, FD_CLOEXEC);
    listen(sock, 10);
    while (1) {
        int fd = accept(sock, 0, 0);
        if (fd < 0) {
            perror("accept");
            break;
        }
        fcntl(fd, F_SETFD, FD_CLOEXEC);
        fcntl(fd, F_SETFL, O_NONBLOCK);
        now = time(0);
        handleRequest(fd);
        close(fd);
    }
    close(sock);
}
