/**
  *  \file planetscentral/talk/talkconnection.cc
  *  \brief Talk: Connection Handling
  */

#include <stdexcept>
#include <ctype.h>
#include <stdio.h>
#include <memory>
#include "planetscentral/talk/talkconnection.h"
#include "cpluslib/misc.h"
#include "int/args.h"
#include "int/array.h"
#include "planetscentral/dbclient/dbsubtree.h"
#include "planetscentral/osutil.h"
#include "planetscentral/talk/cmdfolder.h"
#include "planetscentral/talk/cmdforum.h"
#include "planetscentral/talk/cmdgroup.h"
#include "planetscentral/talk/cmdpm.h"
#include "planetscentral/talk/cmdpost.h"
#include "planetscentral/talk/cmdrender.h"
#include "planetscentral/talk/cmdthread.h"
#include "planetscentral/talk/cmduser.h"
#include "planetscentral/talk/db.h"
#include "planetscentral/talk/talk.h"
#include "planetscentral/talk/user.h"
#include "planetscentral/talk/permission.h"
#include "planetscentral/talk/cmdsyntax.h"

namespace {
    struct Command {
        const char* name;
        void (*func)(TalkConnection&, IntArgBlock&);
    };
}

static const Command commands[] = {
    { "FOLDERLS",      doFolderList },
    { "FOLDERLSPM",    doFolderListPMs },
    { "FOLDERMSTAT",   doFolderMStat },
    { "FOLDERNEW",     doFolderNew },
    { "FOLDERRM",      doFolderRemove },
    { "FOLDERSET",     doFolderSet },
    { "FOLDERSTAT",    doFolderStat },
    { "FORUMADD",      doForumAdd },
    { "FORUMGET",      doForumGet },
    { "FORUMLSPOST",   doForumListPosts },
    { "FORUMLSSTICKY", doForumListStickyThreads },
    { "FORUMLSTHREAD", doForumListThreads },
    { "FORUMMSTAT",    doForumMStat },
    { "FORUMPERMS",    doForumPerms },
    { "FORUMSET",      doForumSet },
    { "FORUMSIZE",     doForumSize },
    { "FORUMSTAT",     doForumStat },
    { "GROUPADD",      doGroupAdd },
    { "GROUPGET",      doGroupGet },
    { "GROUPLS",       doGroupList },
    { "GROUPMSTAT",    doGroupMStat },
    { "GROUPSET",      doGroupSet },
    { "GROUPSTAT",     doGroupStat },
    { "PMCP",          doPMCopy },
    { "PMFLAG",        doPMFlag },
    { "PMMRENDER",     doPMMRender },
    { "PMMSTAT",       doPMMStat },
    { "PMMV",          doPMMove },
    { "PMNEW",         doPMNew },
    { "PMRENDER",      doPMRender },
    { "PMRM",          doPMRemove },
    { "PMSTAT",        doPMStat },
    { "POSTEDIT",      doPostEdit },
    { "POSTGET",       doPostGet },
    { "POSTLSNEW",     doPostListNew },
    { "POSTMRENDER",   doPostMRender },
    { "POSTMSTAT",     doPostMStat },
    { "POSTNEW",       doPostNew },
    { "POSTRENDER",    doPostRender },
    { "POSTREPLY",     doPostReply },
    { "POSTRM",        doPostRemove },
    { "POSTSTAT",      doPostStat },
    { "RENDER",        doRender },
    { "RENDEROPTION",  doRenderOptions },
    { "SYNTAXGET",     doSyntaxGet },
    { "SYNTAXMGET",    doSyntaxMGet },
    { "THREADLSPOST",  doThreadListPosts },
    { "THREADMSTAT",   doThreadMStat },
    { "THREADMV",      doThreadMove },
    { "THREADPERMS",   doThreadPermissions },
    { "THREADRM",      doThreadRemove },
    { "THREADSTAT",    doThreadStat },
    { "THREADSTICKY",  doThreadSticky },
    { "USERLSPOSTED",  doUserListPosted },
    { "USERLSWATCHEDFORUMS", doUserListWatchedForums },
    { "USERLSWATCHEDTHREADS", doUserListWatchedThreads },
    { "USERMARKSEEN",  doUserMarkSeen },
    { "USERNEWSRC",    doUserNewsrc },
    { "USERUNWATCH",   doUserUnwatch },
    { "USERWATCH",     doUserWatch },
};


static bool
isPrintable(const string_t& s)
{
    for (string_t::size_type i = 0; i < s.size(); ++i)
        if (!isalnum(uint8_t(s[i])) && s[i] != '/' && s[i] != '.' && s[i] != '_' && s[i] != '-' && s[i] != ':' && s[i] != ',')
            return false;
    return true;
}

static string_t
getHelp(string_t topic)
{
    if (topic == "GROUP") {
        return "Group commands:\n"
            "GROUPADD <grid> [<key> <value>...]\n"
            "GROUPGET <grid> <key>\n"
            "GROUPLS <grid>\n"
            "GROUPMSTAT <grid>...\n"
            "GROUPSET <grid> [<key> <value>...]\n"
            "GROUPSTAT <grid>\n";
    } else if (topic == "FORUM") {
        return "Forum commands:\n"
            "FORUMADD [<key> <value>...]\n"
            "FORUMGET <fid> <key>\n"
            "FORUMLSPOST <fid> <listoptions>\n"
            "FORUMLSSTICKY <fid> <listoptions>\n"
            "FORUMLSTHREAD <fid> <listoptions>\n"
            "FORUMMSTAT <fid>...\n"
            "FORUMPERMS <fid> <perm>...\n"
            "FORUMSET <fid> <key> <value>...\n"
            "FORUMSIZE <fid>\n"
            "FORUMSTAT <fid>\n";
    } else if (topic == "RENDER") {
        return "Render commands:\n"
            "RENDEROPTION <renderoptions>\n"
            "RENDER <text> <renderoptions>\n";
    } else if (topic == "SYNTAX") {
        return "Syntax commands:\n"
            "SYNTAXGET <key>\n"
            "SYNTAXMGET <keys...>\n";
    } else if (topic == "POST") {
        return "Posting commands:\n"
            "POSTEDIT <mid> <subj> <text>\n"
            "POSTGET <mid> <field>\n"
            "POSTLSNEW <count>\n"
            "POSTMRENDER <mid>...\n"
            "POSTMSTAT <mid>...\n"
            "POSTNEW <fid> <subj> <text> [USER|READPERM|ANSWERPERM <arg>]\n"
            "POSTRENDER <mid> <renderoptions>\n"
            "POSTREPLY <mid> <subj> <text> [USER <arg>]\n"
            "POSTRM <mid>\n"
            "POSTSTAT <mid>\n";
    } else if (topic == "THREAD") {
        return "Thread commands:\n"
            "THREADLSPOST <tid> <listoptions>\n"
            "THREADMSTAT <tid>...\n"
            "THREADMV <tid> <fid>\n"
            "THREADPERMS <tid> <perm>...\n"
            "THREADRM <tid>\n"
            "THREADSTAT <tid>\n"
            "THREADSTICKY <tid> <value>\n";
    } else if (topic == "USER") {
        return "User commands:\n"
            "USERLSPOSTED <uid> []\n"
            "USERLSWATCHEDFORUMS <listoptions>\n"
            "USERLSWATCHEDTHREADS <listoptions>\n"
            "USERMARKSEEN [THREAD n|FORUM n]...\n"
            "USERNEWSRC [GET|SET|CLEAR|ANY|ALL|FIRSTSET|FIRSTCLEAR...]\n"
            "   [POST n n...|RANGE a b|THREAD n|FORUM n]...\n"
            "USERUNWATCH [THREAD n|FORUM n]...\n"
            "USERWATCH [THREAD n|FORUM n]...\n";
    } else if (topic == "FOLDER") {
        return "Folder commands:\n"
            "FOLDERLS\n"
            "FOLDERLSPM <ufid> <listoptions>\n"
            "FOLDERMSTAT <ufid>...\n"
            "FOLDERNEW <name>\n"
            "FOLDERRM <ufid>...\n"
            "FOLDERSET <ufid> <key> <value>...\n"
            "FOLDERSTAT <ufid>\n";
    } else if (topic == "PM") {
        return "PM commands:\n"
            "PMCP <src-ufid> <dst-ufid> <pmid>...\n"
            "PMFLAG <ufid> <clear> <set> <pmid>...\n"
            "PMMRENDER <ufid> <pmid>...\n"
            "PMMSTAT <ufid> <pmid>...\n"
            "PMMV <src-ufid> <dst-ufid> <pmid>...\n"
            "PMNEW <to> <subj> <text> [PARENT <pmid>]\n"
            "PMRENDER <ufid> <pmid> <renderoptions>\n"
            "PMRM <ufid> <pmid>...\n"
            "PMSTAT <ufid> <pmid>\n";
    } else if (topic == "OPTIONS") {
        return "List options (one per command):\n"
            "  CONTAINS <n>\n"
            "  LIMIT <start> <count>\n"
            "  SIZE\n"
            "  SORT <sortkey>\n"
            "Sort keys:\n"
            "  AUTHOR (post, PM)\n"
            "  EDITTIME (post)\n"
            "  FIRSTPOST (thread)\n"
            "  FORUM (thread)\n"
            "  KEY (forum)\n"
            "  LASTPOST (forum, thread)\n"
            "  LASTTIME (forum, thread)\n"
            "  NAME (forum)\n"
            "  SUBJECT (post, PM, thread)\n"
            "  THREAD (post)\n"
            "  TIME (post, PM)\n"
            "Render options (any number):\n"
            "  BASEURL <url>\n"
            "  FORMAT <fmt>\n"
            "Formats:\n"
            "  abstract:FORMAT (out)\n"
            "  break:FORMAT (out)\n"
            "  code:SYNTAX (in)\n"
            "  force:FORMAT (out)\n"
            "  forum (in, out)\n"
            "  html (out)\n"
            "  mail (out)\n"
            "  noquote:FORMAT (out)\n"
            "  quote:FORMAT (out)\n"
            "  raw (out)\n"
            "  text (in, out)\n";
    } else if (topic == "UID") {
        return "User Id specifications (ACL/PM addressee):\n"
            "-<spec>         (ACL) deny these\n"
            "all             (ACL) everyone\n"
            "p:<key>         (ACL) profile key\n"
            "u:<uid>         (ACL/PM) user\n"
            "g:<gid>         (ACL/PM) everyone in game\n"
            "g:<gid>:<slot>  (PM) slot in game\n";
    } else {
        return "Commands:\n"
            "HELP [<topic>]\n"
            "PING\n"
            "USER <uid>\n"
            "FOLDER->\n"
            "FORUM->\n"
            "GROUP->\n"
            "OPTIONS->\n"
            "PM->\n"
            "POST->\n"
            "RENDER->\n"
            "THREAD->\n"
            "UID->\n"
            "USER->\n";
    }
}

/** Constructor. */
TalkConnection::TalkConnection()
    : Connection()
{ }

/** Handle a single request. */
void
TalkConnection::handleRequest()
{
    // Read request
    std::auto_ptr<IntValue> req(readObject(true));
    IntArray* a = dynamic_cast<IntArray*>(req.get());
    if (!a || a->getData()->getDimension(0) == 0) {
        sendError("400 Syntax error");
        return;
    }
    const IntDataSegment& seg = a->getData()->content;

    // Log request
    {
        MutexGuard g(output_mutex);
        if (user.empty()) {
            printf("[cmd]");
        } else {
            printf("[cmd:%s]", user.c_str());
        }
        for (uint32_t i = 0; i < seg.size(); ++i) {
            IntStringValue* sv = dynamic_cast<IntStringValue*>(seg[i]);
            if (sv != 0) {
                if (sv->getValue().empty()) {
                    printf(" ''");
                } else if (sv->getValue().size() < 100 && isPrintable(sv->getValue())) {
                    printf(" %s", sv->getValue().c_str());
                } else {
                    printf(" ...");
                }
            } else {
                printf(" %s", IntValue::toStringOf(seg[i], false).c_str());
            }
        }
        printf("\n");
    }

    // Execute request
    string_t cmd = strUCase(IntValue::toStringOf(seg[0], false));
    IntArgBlock args(seg, 1, seg.size()-1);
    handleCommand(cmd, args);
}

/** Handle a command.
    \param cmd Command
    \param args Arguments */
void
TalkConnection::handleCommand(const string_t& cmd, IntArgBlock& args)
{
    // Command dispatcher
    if (cmd == "PING") {
        // PING
        sendSuccess("PONG");
    } else if (cmd == "HELP") {
        // HELP
        IntStringValue sv(getHelp(strUCase(toStringResult(args.getNext()))));
        sendObject(&sv);
    } else if (cmd == "USER") {
        // USER
        user = toStringResult(args.getNext());
        renderState.setUser(user);
        sendSuccess("OK");
    } else {
        // Other commands via command jump table
        std::size_t i = 0, n = countof(commands);
        while (i < n && cmd != commands[i].name) {
            ++i;
        }
        if (i < n) {
            commands[i].func(*this, args);
        } else {
            sendError("400 Unknown command");
        }
    }
}

/** Check for admin permissions.
    \return presence of admin permissions */
bool
TalkConnection::isAdmin() const
{
    return user.empty();
}

/** Check for admin permissions.
    Throws std::runtime_error if user has none. */
void
TalkConnection::checkAdmin()
{
    if (!isAdmin()) {
        throw std::runtime_error("403 Permission denied");
    }
}

/** Check for user context. */
void
TalkConnection::checkUser()
{
    if (isAdmin()) {
        throw std::runtime_error("403 Must have user context");
    }
}

/** Get user name. */
const string_t&
TalkConnection::getUser() const
{
    return user;
}

/** Check privilege string.
    \param privString Privilege string from database
    \return true iff user is allowed to see the protected item */
bool
TalkConnection::hasPrivilege(string_t privString)
{
    return isAdmin() || checkUserPermission(privString, user);
}

/** Check privilege string. Throws std::runtime_error if user doesn't have permissions.
    \param privString Privilege string from database */
void
TalkConnection::checkPrivilege(string_t privString)
{
    if (!hasPrivilege(privString)) {
        throw std::runtime_error("403 Permission denied");
    }
}

RenderState&
TalkConnection::getRenderState()
{
    return renderState;
}
